#lang racket

(require "CCA.rkt"
         "typing-rules.rkt"
         (prefix-in prototype: "define-judgment-form/define-judgment-form.rkt")
         "ghc-interaction.rkt"
         redex/reduction-semantics)

(provide random-arrow
         main)

(define (main . args)
  (define seed (add1 (random (sub1 (expt 2 31)))))
  (define task #f)
  
  (command-line
   #:argv args
   #:once-each
   ["--seed"
    n
    "Seed with PRG with n"
    (set! seed (string->number n))]
   ["--progress"
    "Prints each tested expression"
    (progress? #t)]
   
   #:once-any
   ["--depth"  
    "Bounds the depth of the typing derivations of random expressions"
    (prototype:bound-type 'depth)]
   ["--size"  
    "Bounds the size of the typing derivations of random expressions"
    (prototype:bound-type 'size)]
   
   #:once-any
   ["--no-revisit-solved"  
    "Disables backtracking to find different solutions to solved goals"
    (prototype:revisit-solved-goals? #f)]
   
   #:once-any
   ["--sample" 
    n s 
    "Show n random arrows with typing derivations of size less than s"
    (set! task (λ () (sample-arrows (string->number n) (string->number s))))]
   ["--test" 
    n s 
    "Tests normalization on n random arrows with typing derivations of size less than s"
    (set! task (λ () (test-normalization (string->number n) (string->number s))))])
  
  (when task
    (printf "Random seed: ~s\n" seed)
    (random-seed seed)
    (task)))

(define progress? (make-parameter #f))

(define (test-normalization attempts size)
  (define cumulative-generation-time 0)
  (define cumulative-checking-time 0)
  
  (time
   (let loop ([remaining attempts])
     (if (zero? remaining)
         (begin
           (printf "No counterexamples in ~s attempts at size ~s\n" attempts size)
           (printf "Generation time: ~ss\n" (round (/ cumulative-generation-time 1000)))
           (printf "Checking time: ~ss\n" (round (/ cumulative-checking-time 1000))))
         (let ([attempt (add1 (- attempts remaining))])
           (when (progress?)
             (printf "Test #~s\n" attempt)
             (printf "Generating...")
             (flush-output (current-output-port)))
           (match-let ([(list expr real) (real-time (random-arrow size))])
             (set! cumulative-generation-time (+ cumulative-generation-time real))
             (when (progress?)
               (printf " ~ss\n" (round (/ real 1000)))
               (pretty-display expr)
               (printf "Checking...")
               (flush-output (current-output-port)))
             (match-let ([(list result real) (real-time (check-normalization expr))])
               (set! cumulative-checking-time (+ cumulative-checking-time real))
               (when (progress?)
                 (printf " ~ss\n\n" (round (/ real 1000))))
               (if (failure? result)
                   (begin
                     (printf "Failure after ~s attempts at size ~s:\n" attempt size)
                     (pretty-display expr)
                     (match result
                       [(raised exn)
                        (displayln "Raised an exception:")
                        (pretty-display exn)]
                       [(abnormal-form norm)
                        (displayln "Result not in CCNF:")
                        (pretty-display norm)]
                       [(type-change norm type)
                        (displayln "Normal form:")
                        (pretty-display norm)
                        (displayln "Expected type:")
                        (pretty-display type)]
                       [(meaning-change norm type)
                        (displayln "Normal form:")
                        (pretty-display norm)
                        (displayln "Type:")
                        (pretty-display type)]))
                   (loop (sub1 remaining))))))))))

(define-syntax-rule (real-time expr)
  (match-let ([(list (list result) _ real _)
               (call-with-values 
                (λ () (time-apply (λ () expr) '()))
                list)])
    (list result real)))

(struct success ())
(struct failure () #:prefab)
(struct raised failure (exn) #:prefab)
(struct abnormal-form failure (norm) #:prefab)
(struct type-change failure (norm type) #:prefab)
(struct meaning-change failure (norm type) #:prefab)

(define (check-normalization e)
  (with-handlers ([exn:fail? raised])
    (match (prototype:generate (typeof+ • ,(to-type-form e) (? τ)) +inf.0)
      [#f (error 'check-normalization "no type for ~s" e)]
      [env (let ([t (random-type-instance (first (dict-ref env 'τ)))]
                 [n (cc-normal e)])
             (if (ccnf? n)
                 (if (prototype:generate (typeof+ • ,(to-type-form n) ,t) +inf.0)
                     (if (check-meaning-preservation e n t)
                         (success)
                         (meaning-change n t))
                     (type-change n t))
                 (abnormal-form n)))])))

(define (check-meaning-preservation e n t)
  (define test-module
    (make-module 
     (term 
      ((orig (haskell-type ,t) (haskell-expr ,e))
       (norm (haskell-type ,t) (haskell-expr ,n))
       (main (IO Bool)
             ((quickCheckWithResult
               (stdArgs "{" maxSize = 35 "," chatty = False "}")
               (\\ (Observation o) -> (o orig) == (o norm)))
              >>=
              return |.| isSuccess))))
     '(Equivalence 
       StreamTransformer
       CCALib
       Test.QuickCheck
       Test.QuickCheck.Test
       "Prelude hiding (init)") 
     '()))
  (define test-result (run-module test-module))
  (delete-file test-module)
  (match test-result
    [(regexp "True\n$") #t]
    [(regexp "False\n$") #f]))

(define (random-expr-instance expr [env (make-hash)])
  (define (random-constant)
    (generate-term CCA constant 1))
  (define (random-number)
    (generate-term CCA real (random 3)))
  (let recur ([e expr])
    (match e
      [`(num ,(prototype:lvar x))
       (instantiate x random-number env)]
      [(prototype:lvar x)
       (instantiate x random-constant env)]
      [(? list?)
       (map recur e)]
      [_ e])))

(define (random-type-instance type [env (make-hash)])
  (define (random-type)
    (generate-term CCA α (random 3)))
  (let recur ([t type])
    (match t
      [(prototype:lvar x) (instantiate x random-type env)]
      [(? list?) (map recur t)]
      [_ t])))

(define (instantiate var make env)
  (match (hash-ref env var #f)
    [#f (define i (make))
        (hash-set! env var i)
        i]
    [i i]))

(define (random-arrow size)
  (cond [(prototype:generate (typeof • (? e) (arr (? α) (? β))) size)
         => (compose from-type-form
                     random-expr-instance
                     second
                     (curry assoc 'e))]
        [else (error 'random-expression "failure at size ~s" size)]))

(define (sample-arrows samples size)
  (time
   (for ([_ (in-range samples)])
     (pretty-display (random-arrow size)))))